/* $Id: metcal.c 37 2009-04-24 19:58:38Z gbv $ */
/***************************************************************************
 *   Copyright (C) 2009 by Guillermo Ballester Valor                       *
 *   gbv@oxixares.com                                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
/*! 
    \file metcal.c 
    \brief file with metcal library routines code

*/
#include "metcal.h"

/*!
  \fn void clean_cal_pol(struct cal_pol *c)
  \brief clean a struct \ref cal_pol
  \param c pointer to the struct to clean 
*/
void clean_cal_pol(struct cal_pol *c)
{
  memset((void *) c, 0, sizeof(struct cal_pol));
}


/*!
  \fn int read_cal_points(struct station_cal *s, const char *file)
  \brief read the calibration points from a file and store them in a cal_pol struct
  \param s pointer to a struct \ref station_cal where all the data are set
  \param file string with the path to the file with the points

  It also computes the spline coeficients. Returns the number of points readed

  A calibration data file are lines with the format:

    Prefix_string  Read_value  Trusted_value

  Prefix_string: a string which label the sensor/parameter

  Read_value:  the value without correction 

  Trusted_value: the corrected value 

  A line begining with '#' is considered a comment.
  It is not necessary to sort the lines. The library will do it.

  EXAMPLE OF CALIBRATION FILE:
----------
# Example of a calibrartion file
# This station has two sensors we designed as 't' and 'h'
# We set 3 points for 't' and 2 for 'h'
h 85 100
t 35.0 34.0
t 12.0 12.8
t 0.0 0.3
h 45 20
---------- 

The way we manage in a calling routine how to organize this points is with the aid
of a struct \ref station_cal . For the above example, we can do someting like

#include "metcal.h"

struct cal_pol CALT, CALH;

struct station_cal CALHT71 = {2, "t", &CALT, "h", &CALH};

*/
int read_cal_points(struct station_cal *s, const char *file)
{
  int i, j, k;
  struct cal_point aux;
  struct cal_pol *c;
  double read, trusted;
  char prefix[32], lin[128];
  FILE *f;

  if ((f = fopen(file,"r")) == NULL)
    {
      fprintf(stderr, gettext("metcal: Cannot open file %s\n"), file);
      return 1;
    }


  while ( fgets(lin, 128, f) != NULL)
    {
      if (lin[0] != '#')
        {

          if (sscanf(lin, "%s %lf %lf", prefix, &read, &trusted) == 3)
            {
              /* search if a valid prefix */
              for (i = 0; i < s->npf; i++)
                {
                  if (strcmp(prefix, s->pf[i].prefix))
                    continue;
                  /* prefix found, set the values */
                  c = s->pf[i].c;
                  c->p[c->np].read = read;
                  c->p[c->np].trusted = trusted;
                  (c->np)++;
                  break;
                }
            }
        }
    }
  fclose (f);

  /* now sort all cal_pol */
  for (k = 0; k < s->npf ; k++)
    {
      c = s->pf[k].c;
      if (c->np <= 1)
        continue;
      /* points in ascending read value */
      for (i = c->np - 1; i > 0; i--)
        {
          for (j = 0; j < i; j++)
            {
              if (c->p[j].read > c->p[j + 1].read)
                {
                  memcpy(&aux, &(c->p[j]), sizeof(struct cal_point));
                  memcpy(&(c->p[j]), &(c->p[j + 1]), sizeof(struct cal_point));
                  memcpy(&(c->p[j + 1]), &aux, sizeof(struct cal_point));
                }
            }
        }
    }

  /* computes spline factors */
  for (k = 0; k < s->npf; k++)
    {
      c = s->pf[k].c;
      if (c->np <= 2)
        continue;
      if (metcal_spline(c))
        return 1;
    }

  /*printf("Leido %d puntos \n", c->np);*/
  return 0;
};


/*!
   \fn double corval( double val, struct cal_pol *c)
   \brief returns the considered good value from a val obtained from sensor
   \param val the input value
   \param c pointer to a struct \ref cal_pol where are set the calibration points for this variable/sensor

   If no calibration points are set it returns directly the input value
*/
double calibrated( double val, struct cal_pol *c)
{
  int i;
  double cor, m;

  if (c == NULL || c->np == 0.0)
    return val;

  if (c->np == 1)
    {
      /* Just one point */
      return (val - c->p[0].read + c->p[0].trusted);
    }
  else if ( c->np == 2)
    {
      /* two points */
      m = (c->p[1].trusted - c->p[0].trusted) / (c->p[1].read - c->p[0].read);
      return ((val - c->p[0].read) * m + c->p[0].trusted);
    }
  else
    {
      /* np > 2 points, cubic spline */
      spline_val(&cor, val, c);
      return cor;
    }
}


/*!
  \fn int metcal_spline(struct cal_pol *c)
  \brief computes natural cubic spline coeficients for a struct \ref cal_pol
  \param c pointer to a struct \ref cal_pol

  This code is based in Numerical Recipes in C
*/
int metcal_spline(struct cal_pol *c)
{
  int i, k, n;
  double p, qn1, sig, un1, *u;

  n = c->np; /* number of points */

  if (n < 3)
    return 1;

  u = (double *) malloc( (n - 1) * sizeof(double));

  /* natural spline */
  u[0] = 0;
  c->p[0].spline = 0;

  /*
  c->p[0].spline = -0.5;
  u[0] = (3.0 / (c->p[1].read - c->p[0].read)) * ((c->p[1].trusted - c->p[0].trusted) / (c->p[1].read  - c->p[0].read) - yp0);
  */

  for (i = 1; i < n - 1 ; i++)
    {
      if (c->p[i].read == c->p[i-1].read)
      {
        fprintf(stderr, gettext("Bad values to routine metcal_spline"));
        free((void *)u);
        return 1;
      }
      sig = (c->p[i].read - c->p[i-1].read) / (c->p[i+1].read - c->p[i-1].read);
      p = sig * c->p[i-1].spline + 2.0;
      c->p[i].spline = (sig - 1.0) / p;
      u[i] = (c->p[i+1].trusted - c->p[i].trusted)/(c->p[i+1].read - c->p[i].read) -
             (c->p[i].trusted - c->p[i-1].trusted)/(c->p[i].read - c->p[i-1].read);
      u[i]= (6.0 * u[i] / (c->p[i+1].read - c->p[i-1].read) - sig * u[i-1]) / p;
    }

  /* natural spline */
  un1 = qn1 = 0.0;

  /*
  qn1 = 0.5;
  un1 = (3.0/(c->p[n-1].read - c->p[n-2].read)) * (ypn1 - (c->p[n-1].trusted - c->p[n-2].trusted) / (c->p[n-1].read - c->p[n-2].read));
  */
  c->p[n-1].spline = (un1 - qn1 * u[n-2])/(qn1 * c->p[n-2].spline + 1.0);

  /* backsubstitution loop in tridiagonal algorithm */
  for (k = n - 2; k >= 0; k--)
    {
      c->p[k].spline = c->p[k].spline * c->p[k + 1].spline + u[k];
    }
  free((void *)u);
  return 0;
}

/*!
  \fn int spline_val(double *y, double x, struct cal_pol *c)
  \brief compute the cubic spline interpolation of a value
  \param y pointer to a double with the interpolated value as output
  \param x de input x value
  \param c pointer to a struct \ref cal_pol where are the data and spline coeficients
 
  The spline coeficients in the struct have to be computed with the aid of \ref metcal_spline
  routine before.
*/
int spline_val(double *y, double x, struct cal_pol *c)
{
  int klo, khi, k, n;
  double h, b, a;

  n = c->np;
  klo = 0;
  khi = n - 1;

  /* find the points */
  while (khi - klo > 1)
    {
      k = (khi + klo) >> 1;
      if (c->p[k].read >  x) khi = k;
      else klo = k;
    }
  h = c->p[khi].read - c->p[klo].read;

  if (h == 0.0)
    {
      fprintf(stderr, ("Bad values to routine spline_val"));
      return 1;
    }

  a = (c->p[khi].read - x) / h;
  b = (x - c->p[klo].read) / h;

  /*Cubic spline polynomial is now evaluated.*/
  *y = a * c->p[klo].trusted + b * c->p[khi].trusted +
       ((a * a * a - a) * c->p[klo].trusted + (b * b * b - b) * c->p[khi].trusted) * (h * h)/6.0;

  return 0;
}


/*
   Saturation pressure of vapor at temperature T (Kelvin) (over water)
   Output is in HPa.
   Input:
   T = Temperature (Kelvin)
   Returns es = the saturation pressure of vapor over water (HPa)
*/
double es(double T)
{
  return ((double)6.112 * exp ((double)17.67 * (T - (double)273.16) / (T - (double)29.65)));
}

/*
    t temperature (Celsius)
    hr Relative humidity
    td dewpoint (celsius)
*/
double td( double t, double hr)
{
    double x, e;

    e = 0.01 * hr * es( t + 273.15);  
    /*x = log(e /6.112);
    t  = (273.15 - 29.65) * x /(17.67 - x);*/
    x = log(e /6.112) / 17.67;
    return (273.16 - 29.65 * x) /(1.0 - x) - 273.15; 
     
}

/*
   Computes wind chill
   wind_chill = 35.74 + 0.6215 * t - 35.75 * v^0.16 + 0.4275* T * v^0.16
   where:
   t = temperature (Farenheit)
   v = vind speed (in knots)

   ARGUMENTS:

   INPUT
    t = temperature (kelvin)
    v = wind speed (in knots)

   OUTPUT
   Returns wind_chill (in Kelvin)

*/
double wind_chill(double t, double v)
{
  double x, auxv, auxt;
  /* pass t to Farenheit */
  auxt = ((t - (double)(273.15)) * (double)(1.8)) + (double)(32.0);
  if (auxt > (double)50.0)
    return t;

  if( v < (double)0.1 )
    auxv = 1.0;
  else
    auxv = exp( 0.16 * log((double)v));
  x = (double)35.74 + (double)0.6215 * auxt - (double)35.75 * auxv  +
      (double)0.4275 * auxt * auxv;
  return ((x - (double)32.0) / (double)1.8 + (double)273.15);
}

/*
        Computes the heat index

        INPUT:
        t = temperature (Kelvin)
        td = dewpoint (kelvin)

        OUTPUT
        returns heat index (kelvin)
*/

#define HI_FORMULA_ALT
double heat_index( double t, double td)
{
  double xt, hr, fh, xt2, hr2;

  if (t <= (double)(292.0))
    return t;

  /*hr = (double)0.01 * humrel( t, td);*/
  hr = 100.0*es(td)/es(t);
  xt = (t - (double)(273.15)) * (double)(1.8) + (double)(32.0);

  /* some util powers */
  xt2 = xt * xt;
  hr2 = hr * hr;
  /*printf("%lf %lf %lf\n", t, td, hr);*/

  /* Heat Index */
#if defined(HI_FORMULA_ALT)

  fh = (double)(-42.379) + ((double)(2.04901523) * xt)
       + ((double)(10.14333127) * hr)
       - ((double)(0.22475541) * xt * hr)
       - ((double)(6.83783e-3) * xt2)
       - ((double)(5.481717e-2) * hr2)
       + ((double)(1.22874e-3) * xt2 * hr)
       + ((double)(8.5282e-4) * xt * hr2)
       - ((double)(1.99e-6) * xt2 * hr2);
  /* no hace nada util salvo arreglar un maldito bug en gcc */
  xt = es((t + td)* 0.5);
  /*printf("hi=%lf\n", fh);*/
#else

  {
    double xt3, hr3;
    xt3 = xt2 * xt;
    hr3 = hr2 * hr;
    fh = 16.923 + 1.85212 * 0.1 * xt + 5.37941 * hr - 1.00254 * 0.1 * xt * hr
         + 9.41695 * 0.001 * xt2 + 7.28898 * 0.001 * hr2
         + 3.45372 * 0.0001 * xt2 * hr - 8.14971 * 0.0001 * xt *hr2
         + 1.02102 * 0.00001 * xt2 * hr2 - 3.8646 * 0.00001 * xt3
         + 2.91583 * 0.00001 * hr3 + 1.42721 * 0.000001 * xt3 * hr
         + 1.97483 * 0.0000001 * xt * hr3 - 2.18429 * 0.00000001 * xt3 * hr2
         + 8.43296 * 0.0000000001 * xt2 * hr3 - 4.81975 * 0.00000000001 * xt3 * hr3;
  }
#endif
  /*printf("%lf %lf %lf %lf\n", t, td, hr, fh);*/
  xt = ((fh - (double)(32.0))/(double)(1.8) + (double)(273.15)) ;
  return xt;
}
