/* $Id: wwsr.h 38 2009-04-25 21:35:47Z gbv $ */
/***************************************************************************
 *                                                                         *
 *   wwsr - Wireless Weather Station Reader                                *
 *   2007 dec 19, Michael Pendec (michael.pendec@gmail.com)                *
 *   Version 0.1                                                           *
 *   2008 jan 24 Svend Skafte (svend@skafte.net)                           *
 *   Modifications for different firmware version(?)                       *
 *   2009 Jan 21 Guillermo Ballester gbv@oxixares.com                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
/*! \file wwsr.h
    \brief include file for wwsr
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* To active Berkeley DB support */
/*#define WITH_DB*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <signal.h>
#include <stdint.h>
#include <ctype.h>
#include <time.h>
#include <usb.h>
#include <libintl.h>
#ifdef WITH_DB
# include "metdb.h"
#endif

/* library for correction */
#include "metcal.h"

/* Masks to select parameters */
#define MASK_HUM_INT 1
#define MASK_HUM_EXT 2
#define MASK_TEMP_INT 4
#define MASK_TEMP_EXT 8
#define MASK_PRES 16
#define MASK_WSPD 32
#define MASK_WDIR 64
#define MASK_RAIN 128
#define MASK_INTERV 256
#define MASK_DATIME 512
#define MASK_WIND (MASK_WSPD | MASK_WDIR)
#define MASK_HUM (MASK_HUM_INT | MASK_HUM_EXT)
#define MASK_TEMP (MASK_TEMP_INT | MASK_TEMP_EXT)
#define MASK_ALL (MASK_HUM | MASK_TEMP | MASK_PRES | MASK_WIND | MASK_RAIN | MASK_INTERV)
#define MASK_PREC MASK_RAIN
#define PATH_LEN  256

/* size of station memory */
#define MEMSIZE 0x10000

/*!
  \struct obs_data 
  \brief struct to read a data sample and store it in integer format
*/
struct obs_data 
{
   time_t datime; /*!< unix time of observation (round to minute) */
   uint16_t interval; /*!< minutes since prior data */
   uint16_t hum_int; /*!< internal humidity in % */
   int16_t temp_int; /*!< Interior temperature in thents of celsius */
   int16_t hum_out; /*!< external humidity in % */
   int16_t temp_out; /*!< external temperature in thents of celsius */
   uint16_t press; /*!< pressure in thenths of Hectopascal */
   uint16_t wspd; /*!< wind speed in thents of m/s */
   uint16_t wgust; /*!< wind gust in thents of m/s */
   uint16_t wdir; /*!< windir in directions */
   uint16_t prec; /*!< accumulated precipitation in 0'3 mm units */
   uint8_t mask; /*!< it is FF if no data is present */
   uint8_t buf[16]; /*!< the buffer src from wwsr */
};

/*!
  \struct real_data 
  \brief struct to store calibrated data in float form
*/
struct real_data 
{
   int mask; /*!< bit mask to ressume what valid information the struct have */
   time_t datime; /*!< unix time of observation (round to minute) */
   double interval; /*!< seconds since prior data */
   double hum_int; /*!< internal humidity in % */
   double temp_int; /*!< Interior temperature (celsius) */
   double tdew_int; /*!< Interior dewpoint temperature (celsius) */
   double hum_out; /*!< external humidity in % */
   double temp_out; /*!< external temperature (celsius), computed */
   double tdew_out; /*!< external dewpoint temperature (celsius), computed */
   double press; /*!< pressure Hectopascal */
   double wspd; /*!< wind speed (m/s) */
   double wgust; /*!< wind gust (m/s) */
   double wdir; /*!< windir in directions */
   double prec; /*!< accumulated precipitation (mm) */
};

struct stat_var{
   size_t n; /*!< Number of data*/
   size_t tiempo; /*!< Amount of seconds with data */
   double vmax;  /*!< Max value */
   time_t tvmax; /*!< unix time of max valur */
   double vmin;  /*!< min value */
   time_t tvmin;  /*!< unix time of min value */
   double acum; /*!< Amount / mean of data */
};

struct stat_summary {
   int mask;
   size_t n;
   time_t begin;
   time_t end;
   struct stat_var hum_int; /*!< stat resul for interior humidity */
   struct stat_var temp_int; /*!< stat resul for interior temperature */
   struct stat_var tdew_int; /*!< stat resul for interior dewpoint temperature */
   struct stat_var hum_out; /*!< stat resul for exterior humidity */
   struct stat_var temp_out; /*!< stat resul for exterior temperature */
   struct stat_var tdew_out; /*!< stat resul for exterior dewpoint temperature */
   struct stat_var press; /*!< stat resul for pressure */
   struct stat_var wspd; /*!< stat resul for wind speed */
   struct stat_var wgust; /*!< stat resul for wind speed */
   struct stat_var prec; /*!< stat resul for accumulated precipitation */
};


struct extrem_data 
{
   double dat;
   struct tm tim;
};

extern int VIEWMASK;
extern struct usb_dev_handle *devh;
extern int ret, mempos, resetws, pdebug;
extern unsigned char    LOGBUF[MEMSIZE], CURBUF[32];
extern char usbmsg[16], *endptr;
extern unsigned char    buf2[400];
extern char wd[17][4];
extern struct extrem_data hin_max, hin_min, tin_max, tin_min, hout_max, hout_min, tout_max, tout_min, chill_max,
          chill_min, dwp_max, dwp_min, apres_max, apres_min, rpres_max, rpres_min, wsp_max, gust_max, 
          rh_max, rd_max, rw_max, rm_max, rt_max; 
extern char BACKUPFILE[PATH_LEN], INPUTFILE[PATH_LEN], CALFILE[PATH_LEN];
extern char DATEBEGIN[16], DATEEND[16];
extern int ONLYLAST, FROMFILE, EXTREM, NOCAL, LASTN, GMTOFF, TIMESET;
extern time_t TBEGIN, TEND;
extern FILE *F;
extern struct cal_pol CALti, CALto, CALhi, CALho, CALp, CALw, CALr;
extern struct station_cal CALWWSR;

#ifdef WITH_DB
extern int OVERWRITE, FROMDB, KILLDATA, CLIMAT, INTERP;
extern char DBFILE[PATH_LEN];
extern size_t INSERTED, REJECTED, ERRORS;
extern struct stat_summary Summary;
#endif

void _close_readw();
struct usb_device *find_device(int vendor, int product);
void print_bytes(char *bytes, int len);
void _open_readw();
void _init_wread();
void _send_usb_msg( char msg1[1],char msg2[1],char msg3[1],char msg4[1],char msg5[1],char msg6[1],char msg7[1],char msg8[1] );
void _read_usb_msg(char *buffer);

void reset_ws(void);
size_t get_main_block(char *buf2);
int buf2obs (struct obs_data *o, unsigned char *buf);
void print_obs_verbose(FILE *f, struct obs_data *o);
void print_obs(FILE *f, struct obs_data *o, char *sep);
void print_obs_nores(FILE *f, struct obs_data *o, char *sep);
uint8_t bcdval( uint8_t a);
void read_date(struct tm *tim, uint8_t *bf);
int read_arguments(int argc, char **argv);
uint16_t read_uint16(uint8_t *bf);
int16_t read_int16(uint8_t *bf);
void print_extreme(FILE *f, struct extrem_data *d, char *name, char *unit);
void print_all_extremes(FILE *f);
int write_backup_file(char *name, size_t nbytes);
size_t read_station_buf( uint8_t *b , size_t limit);
double value( double x, struct cal_pol *c);
unsigned int get_pos(uint16_t current, int nobs);
size_t read_current(uint8_t *b, uint16_t current);
void read_extremes(uint8_t *bf);
int obsdata_to_realdata(struct real_data *r, struct obs_data *o);
void print_stat_summary(struct stat_summary *s, FILE *f);
int summarize_realdata(struct stat_summary *s, const struct real_data *r);
int interp_data(struct real_data *r, struct obs_data *o, size_t nobs, time_t datime);
void print_real_data(FILE *f, struct real_data *r,const char *sep);



#ifdef WITH_DB

int insert_db_obs(DB * dbp, struct obs_data *obs);
int get_db_obs(DB *dbp, struct obs_data *obs, size_t maxobs, time_t tbegin, time_t tend);
int get_db_obs_wide(DB *dbp, struct obs_data *obs, size_t maxobs, time_t tbegin, time_t tend);
int del_db_obs(DB *dbp, uint8_t mask, time_t tbegin, time_t tend);

#endif



