/* $Id: wwsr_stats.c 38 2009-04-25 21:35:47Z gbv $ */
/***************************************************************************
 *   Copyright (C) 2009 by Guillermo Ballester Valor                       *
 *   gbv@oxixares.com                                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
/*!
  \file wwsr_stats.c
  \brief This file include functions to extract climatic/statistics from a DB file with wwsr data
*/
#include "wwsr.h"

/*!
  \fn int obsdata_to_realdata(struct real_data *r, struct obs_data *o)
  \brief convert the data in a struct \ref obs_data and pass them to a struct \ref real_data
  \param r pointer to the target \ref real_data
  \param o pointer to source struct \ref obs_data

  Returns 1 if succeded, 0 otherwise
*/
int obsdata_to_realdata(struct real_data *r, struct obs_data *o)
{
  /* check if valid data */
  if ( o->mask == 0xFF || (o->hum_int == 0 && o->interval == 0))
    return 0;

  r->datime = o->datime;
  if (r->datime > 0)
    r->mask |= MASK_DATIME;

  r->interval = 60.0 * (unsigned int) o->interval;
  if (r->interval > 0.0)
    r->mask |= MASK_INTERV;

  r->hum_int = value((double) o->hum_int, &CALhi);
  if (r->hum_int > 5.0 && r->hum_int <= 100.0)
    r->mask |= MASK_HUM_INT;

  r->temp_int = value((double)(o->temp_int) * 0.1, &CALti);
  if (r->temp_int > -15.0 && r->temp_int <= 50.0)
    r->mask |= MASK_TEMP_INT;

  if ((r->mask & MASK_TEMP_INT) && (r->mask & MASK_HUM_INT))
    {
      r->tdew_int = td(r->temp_int, r->hum_int);
    }

  r->hum_out = value((double) o->hum_out, &CALho);
  if (r->hum_out > 5.0 && r->hum_out <= 100.0)
    r->mask |= MASK_HUM_EXT;

  r->temp_out = value((double)(o->temp_out) * 0.1, &CALto);
  if (r->temp_out > -35.0 && r->temp_out <= 50.0)
    r->mask |= MASK_TEMP_EXT;

  if ((r->mask & MASK_TEMP_EXT) && (r->mask & MASK_HUM_EXT))
    {
      r->tdew_out = td(r->temp_out, r->hum_out);
    }

  r->press = value((double)(o->press) * 0.1, &CALp);
  if (r->press < 1055.0 && r->press > 700.0)
    r->mask |= MASK_PRES;

  r->wspd = value((double)(o->wspd) * 0.1, &CALw);
  r->wgust = value((double)(o->wgust) * 0.1, &CALw);
  if (r->wspd >= 0.0 && r->wspd < 200.0 && r->wgust >= 0.0 && r->wgust < 200.0)
    r->mask |= MASK_WSPD;

  if (o->wdir > 16)
    r->wdir = 999.9;
  else
    r->wdir = 22.5 * (double) o->wdir;
  if (r->wdir == 0.0)
    r->wdir = 360.0;

  if (r->wdir >= 0.0 && r->wdir <= 360.0)
    r->mask |= MASK_WDIR;

  r->prec = value((double)(o->prec) * 0.3, &CALr);
  if (r->prec >= 0.0 && r->prec < 200.0)
    r->mask |= MASK_RAIN;
  return 1;
}


/*!
  \fn int summarize_realdata(struct stat_summary *s, const struct real_data *r)
  \brief Computes a struct \ref real_data into a struct \ref stat_summary
  \param s pointer to the target struct \ref stat_summary
  \param r pointer to a struct \ref real_summary with the data to be taken account

  If any data has been computed, it returned 1, otherwise 0.
*/
int summarize_realdata(struct stat_summary *s, const struct real_data *r)
{
  double aux, interval;

  // Date and time must be in the interval
  if ((r->mask & MASK_DATIME) == 0)
    return 0;

  if ((r->datime > s->end) || (r->datime < s->begin))
    return 0;

  if (r->datime == s->begin || r->datime == s->end)
    interval = r->interval * 0.5;
  else
    interval = r->interval;

  // hum_int
  if (r->mask & MASK_HUM_INT)
    {
      s->hum_int.tiempo += interval;
      s->hum_int.acum += r->hum_int * interval;
      if (s->hum_int.n == 0 || s->hum_int.vmax < r->hum_int)
        {
          s->hum_int.vmax = r->hum_int;
          s->hum_int.tvmax = r->datime;
        }
      if (s->hum_int.n == 0 || s->hum_int.vmin > r->hum_int)
        {
          s->hum_int.vmin = r->hum_int;
          s->hum_int.tvmin = r->datime;
        }
      (s->hum_int.n)++;
    }

  // temp_int
  if (r->mask & MASK_TEMP_INT)
    {
      s->temp_int.tiempo += interval;
      s->temp_int.acum += (r->temp_int * interval);
      if (s->temp_int.n == 0 || s->temp_int.vmax < r->temp_int)
        {
          s->temp_int.vmax = r->temp_int;
          s->temp_int.tvmax = r->datime;
        }
      if (s->temp_int.n == 0 || s->temp_int.vmin > r->temp_int)
        {
          s->temp_int.vmin = r->temp_int;
          s->temp_int.tvmin = r->datime;
        }
      (s->temp_int.n)++;
    }

  // tdew_int
  if ((r->mask & MASK_TEMP_INT) && (r->mask & MASK_HUM_INT))
    {
      s->tdew_int.tiempo += interval;
      s->tdew_int.acum += (r->tdew_int * interval);
      if (s->tdew_int.n == 0 || s->tdew_int.vmax < r->tdew_int)
        {
          s->tdew_int.vmax = r->tdew_int;
          s->tdew_int.tvmax = r->datime;
        }
      if (s->tdew_int.n == 0 || s->tdew_int.vmin > r->tdew_int)
        {
          s->tdew_int.vmin = r->tdew_int;
          s->tdew_int.tvmin = r->datime;
        }
      (s->tdew_int.n)++;
    }

  // hum_out
  if (r->mask & MASK_HUM_EXT)
    {
      s->hum_out.tiempo += interval;
      s->hum_out.acum += (r->hum_out * interval);
      if (s->hum_out.n == 0 || s->hum_out.vmax < r->hum_out)
        {
          s->hum_out.vmax = r->hum_out;
          s->hum_out.tvmax = r->datime;
        }
      if (s->hum_out.n == 0 || s->hum_out.vmin > r->hum_out)
        {
          s->hum_out.vmin = r->hum_out;
          s->hum_out.tvmin = r->datime;
        }
      (s->hum_out.n)++;
    }

  // temp_out
  if (r->mask & MASK_TEMP_EXT)
    {
      s->temp_out.tiempo += interval;
      s->temp_out.acum += (r->temp_out * interval);
      if (s->temp_out.n == 0 || s->temp_out.vmax < r->temp_out)
        {
          s->temp_out.vmax = r->temp_out;
          s->temp_out.tvmax = r->datime;
        }
      if (s->temp_out.n == 0 || s->temp_out.vmin > r->temp_out)
        {
          s->temp_out.vmin = r->temp_out;
          s->temp_out.tvmin = r->datime;
        }
      (s->temp_out.n)++;
    }

  // tdew_out
  if ((r->mask & MASK_TEMP_EXT) && (r->mask & MASK_HUM_EXT))
    {
      s->tdew_out.tiempo += interval;
      s->tdew_out.acum += (r->tdew_out * interval);
      if (s->tdew_out.n == 0 || s->tdew_out.vmax < r->tdew_out)
        {
          s->tdew_out.vmax = r->tdew_out;
          s->tdew_out.tvmax = r->datime;
        }
      if (s->tdew_out.n == 0 || s->tdew_out.vmin > r->tdew_out)
        {
          s->tdew_out.vmin = r->tdew_out;
          s->tdew_out.tvmin = r->datime;
        }
      (s->tdew_out.n)++;
    }

  // press
  if (r->mask & MASK_PRES)
    {
      s->press.tiempo += interval;
      s->press.acum += (r->press * interval);
      if (s->press.n == 0 || s->press.vmax < r->press)
        {
          s->press.vmax = r->press;
          s->press.tvmax = r->datime;
        }
      if (s->press.n == 0 || s->press.vmin > r->press)
        {
          s->press.vmin = r->press;
          s->press.tvmin = r->datime;
        }
      (s->press.n)++;
    }

  // wspd
  if (r->mask & MASK_WSPD)
    {
      s->wspd.tiempo += interval;
      s->wspd.acum += (r->wspd * interval);
      if (s->wspd.n == 0 || s->wspd.vmax < r->wspd)
        {
          s->wspd.vmax = r->wspd;
          s->wspd.tvmax = r->datime;
        }
      if (s->wspd.n == 0 || s->wspd.vmin > r->wspd)
        {
          s->wspd.vmin = r->wspd;
          s->wspd.tvmin = r->datime;
        }
      (s->wspd.n)++;

      s->wgust.tiempo += interval;
      s->wgust.acum += (r->wgust * interval);
      if (s->wgust.n == 0 || s->wgust.vmax < r->wgust)
        {
          s->wgust.vmax = r->wgust;
          s->wgust.tvmax = r->datime;
        }
      if (s->wgust.n == 0 || s->wgust.vmin > r->wgust)
        {
          s->wgust.vmin = r->wgust;
          s->wgust.tvmin = r->datime;
        }
      (s->wgust.n)++;

    }


  /* Precipitation is different
    in acum we store the acumulate prec of interval
    in vmax we store the maximum intensity (mm/hour)
    in vmin we store the last prec
  */
  if (r->mask & MASK_RAIN)
    {
      if (s->prec.n == 0)
        {
          // first value
          s->prec.acum = 0.0;
          s->prec.vmin = r->prec;
          s->prec.vmax = 0.0;
          (s->prec.n)++;
        }
      else
        {
          aux = r->prec - s->prec.vmin;
          if (aux >= 0.0)
            {
              s->prec.acum += aux;
              if (aux > 0.0 && (3600.0 * aux / r->interval) > s->prec.vmax)
                {
                  s->prec.vmax =  3600.0 * aux / r->interval;
                  s->prec.tvmax = r->datime;
                }
              s->prec.vmin = r->prec;
            }
          s->prec.tiempo += r->interval;
          (s->prec.n)++;
        }
    }

  // at the end update global mask and n
  s->mask |= r->mask;
  (s->n)++;

  return 1;
}

void print_stat_summary(struct stat_summary *s, FILE *f)
{
  struct tm t;
  char aux[32];

  if (s->mask & MASK_TEMP_EXT)
    {
      gmtime_r(&(s->temp_out.tvmax), &t);
      strftime(aux, 32, "%H:%M %d/%m/%Y", &t);
      fprintf( f, "TMAX =%6.1lfC (%s). ", s->temp_out.vmax, aux);
      gmtime_r(&(s->temp_out.tvmin), &t);
      strftime(aux, 32, "%H:%M %d/%m/%Y", &t);
      fprintf( f, "TMIN =%6.1lfC (%s). ", s->temp_out.vmin, aux);
      fprintf( f, "TMED =%6.1lfC ", s->temp_out.acum / s->temp_out.tiempo);
      fprintf( f, "VALID=%.1lf%%\n", 100.0 * s->temp_out.tiempo /(s->end - s->begin));
    }
  if (s->mask & MASK_HUM_EXT)
    {
      gmtime_r(&(s->hum_out.tvmax), &t);
      strftime(aux, 32, "%H:%M %d/%m/%Y", &t);
      fprintf( f, "HRMAX=%6.1lf%% (%s). ", s->hum_out.vmax, aux);
      gmtime_r(&(s->hum_out.tvmin), &t);
      strftime(aux, 32, "%H:%M %d/%m/%Y", &t);
      fprintf( f, "HRMIN=%6.1lf%% (%s). ", s->hum_out.vmin, aux);
      fprintf( f, "HRMED=%6.1lf%% ", s->hum_out.acum / s->hum_out.tiempo);
      fprintf( f, "VALID=%.1lf%%\n", 100.0 * s->hum_out.tiempo /(s->end - s->begin));
    }
  if ((s->mask & MASK_HUM_EXT) && (s->mask & MASK_TEMP_EXT))
    {
      gmtime_r(&(s->tdew_out.tvmax), &t);
      strftime(aux, 32, "%H:%M %d/%m/%Y", &t);
      fprintf( f, "TDMAX=%6.1lfC (%s). ", s->tdew_out.vmax, aux);
      gmtime_r(&(s->tdew_out.tvmin), &t);
      strftime(aux, 32, "%H:%M %d/%m/%Y", &t);
      fprintf( f, "TDMIN=%6.1lfC (%s). ", s->tdew_out.vmin, aux);
      fprintf( f, "TDMED=%6.1lfC ", s->tdew_out.acum / s->tdew_out.tiempo);
      fprintf( f, "VALID=%.1lf%%\n", 100.0 * s->tdew_out.tiempo / (s->end - s->begin));
    }
  if ((s->mask & MASK_WSPD))
    {
      gmtime_r(&(s->wgust.tvmax), &t);
      strftime(aux, 32, "%H:%M %d/%m/%Y", &t);
      fprintf( f, "GUST =%6.1lfm/s (%s). ", s->wgust.vmax, aux);
      fprintf( f, "VMED =%6.1lfm/s ", s->wspd.acum / s->wspd.tiempo);
      fprintf( f, "VALID=%.1lf%%\n", 100.0 * s->wspd.tiempo / (s->end - s->begin));
    }
  if (s->mask & MASK_PRES)
  {
      gmtime_r(&(s->press.tvmax), &t);
      strftime(aux, 32, "%H:%M %d/%m/%Y", &t);
      fprintf( f, "PMAX =%6.1lfHp (%s). ", s->press.vmax, aux);
      gmtime_r(&(s->press.tvmin), &t);
      strftime(aux, 32, "%H:%M %d/%m/%Y", &t);
      fprintf( f, "PMIN =%6.1lfHp (%s). ", s->press.vmin, aux);
      fprintf( f, "PMED =%6.1lfHp ", s->press.acum / s->press.tiempo);
      fprintf( f, "VALID=%.1lf%%\n", 100.0 * s->press.tiempo /(s->end - s->begin));
  }
  if (s->mask & MASK_PREC)
    {
      fprintf( f, "PREC =%6.1lf mm. ", s->prec.acum);
      gmtime_r(&(s->prec.tvmax), &t);
      strftime(aux, 32, "%H:%M %d/%m/%Y", &t);
      fprintf( f, "INT MAX.=%6.1lf mm/hr (%s) ", s->prec.vmax, aux);
      fprintf( f, "VALID=%.1lf%%\n", 100.0 * s->prec.tiempo / (s->end - s->begin));
    }
  
}


int interp_data(struct real_data *r, struct obs_data *o, size_t nobs, time_t datime)
{
  double tx;
  size_t i = 0;
  struct real_data r1, r2;

  while (o[i].datime < datime && i < nobs) i++;

  if (i < nobs && o[i].datime == datime)
    {
      obsdata_to_realdata( r, &o[i]);
      return 1;
    }

  if (i < nobs && i > 0)
    {
      memset(&r1, 0, sizeof(struct real_data));
      memset(&r2, 0, sizeof(struct real_data));
      obsdata_to_realdata( &r1, &o[i - 1]);
      obsdata_to_realdata( &r2, &o[i]);

      r->mask = (r1.mask & r2.mask);
      if ((r->mask & MASK_DATIME) == 0)
        return 0;
      tx = ((double)datime - r1.datime) / ((double)r2.datime - r1.datime);
      r->datime = datime;
      r->interval = datime - r1.datime;

      if (r->mask & MASK_HUM_INT)
        r->hum_int = r1.hum_int + (r2.hum_int - r1.hum_int) * tx;

      if (r->mask & MASK_HUM_EXT)
        r->hum_out = r1.hum_out + (r2.hum_out - r1.hum_out) * tx;

      if (r->mask & MASK_TEMP_INT)
        r->temp_int = r1.temp_int + (r2.temp_int - r1.temp_int) * tx;

      if (r->mask & MASK_TEMP_EXT)
        r->temp_out = r1.temp_out + (r2.temp_out - r1.temp_out) * tx;

      if ((r->mask & MASK_HUM_INT) && (r->mask & MASK_TEMP_INT))
        r->tdew_int = r1.tdew_int + (r2.tdew_int - r1.tdew_int) * tx;

      if ((r->mask & MASK_HUM_INT) && (r->mask & MASK_TEMP_INT))
        r->tdew_out = r1.tdew_out + (r2.tdew_out - r1.tdew_out) * tx;

      if (r->mask & MASK_PRES)
        r->press = r1.press + (r2.press - r1.press) * tx;

      if (r->mask & MASK_WSPD)
        {
          r->wspd = r1.wspd + (r2.wspd - r1.wspd) * tx;
          r->wgust = r1.wgust + (r2.wgust - r1.wgust) * tx;
        }

      if (r->mask & MASK_WDIR)
        r->wdir = r1.wdir + (r2.wdir - r1.wdir) * tx;

      if (r->mask & MASK_RAIN)
        r->prec = r1.prec + (r2.prec - r1.prec) * tx;
    }
  else
    return 0;
  return 1;
}
